# Author: Scott Woods <kipjak.baker@gmail.com>
# MIT License
#
# Copyright (c) 2017-2023 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Management of a pool of threads.

.
"""
__docformat__ = 'restructuredtext'

from collections import deque

from .general_purpose import *
from .virtual_memory import *
from .convert_memory import *
from .message_memory import *
from .convert_signature import *
from .convert_type import *
from .virtual_runtime import *
from .virtual_point import *
from .point_runtime import *
from .routine_point import *
from .virtual_codec import *
from .point_machine import *
from .bind_type import *
from .object_collector import *
from .object_directory import *
from .process_object import *
from .get_response import *

__all__ = [
	'JoinSpool',
	'LeaveSpool',
	'ObjectSpool',
]

#
#
class JoinSpool(object):
	def __init__(self, worker_address: Address=None, role_name: str=None):
		self.worker_address = worker_address
		self.role_name = role_name

class LeaveSpool(object):
	def __init__(self, worker_address: Address=None, role_name: str=None):
		self.worker_address = worker_address
		self.role_name = role_name

bind(JoinSpool)
bind(LeaveSpool)

class INITIAL: pass
class PENDING: pass
class EXECUTING: pass
class CLEARING: pass
class SPOOLING: pass

SPOOL_SPAN = 32

class ObjectSpool(Point, StateMachine):
	"""
	Distribute messages across a pool of computing resources.

	:param object_type: type of asynchronous object
	:type object_type: :ref:`object type<kj-object-type>`
	:param args: positional arguments to pass on object creation
	:param role_name: name of process objects, e.g. spool-{i} or None
	:param object_count: number of objects to create
	:param size_of_queue: maximum number of pending messages
	:param responsiveness: expected performance before imposing busy state
	:param busy_pass_rate: rate of messages processed in busy state, as a denominator
	:param stand_down: delay in seconds before restart of terminated object
	:param settings: named arguments to pass on object creation
	"""
	def __init__(self, object_type, *args, role_name: str=None,
			object_count: int=8, size_of_queue: int=None,
			responsiveness: float=5.0, busy_pass_rate: int=10, stand_down: float=30.0, **settings):
		Point.__init__(self)
		StateMachine.__init__(self, INITIAL)
		self.object_type = object_type
		self.args = args
		self.role_name = role_name
		self.object_count = object_count
		self.size_of_queue = size_of_queue
		self.responsiveness = responsiveness
		self.busy_pass_rate = busy_pass_rate
		self.stand_down = stand_down
		self.settings = settings

		self.idle_object = deque()
		self.pending_request = deque()
		self.working_object = {}
		self.span = deque()
		self.total_span = 0.0
		self.average = 0.0
		self.shard = 0

		if responsiveness is None:
			self.no_response = None
		else:
			self.no_response = responsiveness * 5.0

	def submit_request(self, message, forward_response, return_address, presented):
		if self.responsiveness is None:
			pass
		elif self.average < self.responsiveness:
			pass
		else:
			self.shard += 1
			if self.shard % self.busy_pass_rate:
				self.send(Busy(f'message rejected by spool (average response time {self.average:.2f})'), return_address)
				return

		idle = self.idle_object.popleft()
		r = self.create(GetResponse, message, idle, seconds=self.no_response)
		self.working_object[idle] = r
		self.on_return(r, forward_response, idle=idle, return_address=return_address, started=presented)

def ObjectSpool_INITIAL_Start(self, message):
	oc = self.object_count
	sos = self.size_of_queue
	r = self.responsiveness
	sd = self.stand_down

	if oc < 1 or (sos is not None and sos < 1) or (r is not None and r < 0.5) or (sd is not None and sd < 2.0):
		self.complete(Faulted(f'unexpected parameters (count={oc}, size={sos}, responsiveness={r}), stand_down={sd})'))

	if self.object_type is None:
		return SPOOLING

	role_name = self.role_name

	for i in range(oc):
		if self.object_type == ProcessObject:
			if not role_name:
				role_name = 'spool-{i}'
			r = role_name.format(i=i)
			a = self.create(self.object_type, *self.args, role_name=r, **self.settings)
		else:
			r = i
			a = self.create(self.object_type, *self.args, **self.settings)
		self.assign(a, r)
		self.idle_object.append(a)

	return SPOOLING

def forward_response(self, value, kv):
	# Completion of a request/responsesequence.
	# Record the idle process.
	self.working_object.pop(kv.idle, None)
	self.idle_object.append(kv.idle)

	# Update the performance metric. Dont include
	# timeouts as they happen for reasons like dropped
	# connections and skew the stats for a long time.
	if not isinstance(value, TimedOut):
		span = clock_now() - kv.started
		self.total_span += span
		self.span.append(span)
		while len(self.span) > SPOOL_SPAN:
			s = self.span.popleft()
			self.total_span -= s
		self.average = self.total_span / len(self.span)

	# Deliver reponse to the original client.
	m = cast_to(value, self.returned_type)
	self.send(m, kv.return_address)
	if not self.pending_request:
		return
	message, return_address, presented = self.pending_request.popleft()

	# There is a request-to-go and an available process.
	self.submit_request(message, forward_response, return_address, presented)

def ObjectSpool_SPOOLING_JoinSpool(self, message):
	self.idle_object.append(message.worker_address)
	if not self.pending_request:
		return SPOOLING
	message, return_address, presented = self.pending_request.popleft()

	# There is a request-to-go and an available process.
	self.submit_request(message, forward_response, return_address, presented)
	return SPOOLING

def ObjectSpool_SPOOLING_LeaveSpool(self, message):
	a = message.worker_address

	# Its either idle, or its waiting on
	# an active request.
	try:
		self.idle_object.remove(a)
		return SPOOLING
	except ValueError:
		pass

	p = self.working_object.pop(a, None)
	if p is not None:
		self.send(Stop(), p)
	return SPOOLING

def ObjectSpool_SPOOLING_Unknown(self, message):
	m = cast_to(message, self.received_type)
	t = clock_now()
	if not self.idle_object:
		if self.object_type is None and not self.working():
			# Object_type is None -> no callbacks for worker restart.
			# No idle workers, external discovery of workers and nothing underway.
			text = f'Service temporarily unavailable (no worker and no immediate prospects of one)'
			self.reply(TemporarilyUnavailable(text=text))
			return SPOOLING
		len_pending = len(self.pending_request)
		if self.size_of_queue is None or len_pending < self.size_of_queue:
			self.pending_request.append((m, self.return_address, t))
			return SPOOLING
		self.reply(Overloaded(f'message rejected by spool (overloaded, {len_pending} pending)'))
		return SPOOLING

	# There is a request-to-go and an available process.
	self.submit_request(m, forward_response, self.return_address, t)
	return SPOOLING

def ObjectSpool_SPOOLING_Returned(self, message):
	d = self.debrief()
	if isinstance(d, OnReturned):
		d(self, message)
		return SPOOLING

	if self.object_type is None:
		return SPOOLING

	self.trace(f'Spool process termination', returned_value=message_to_tag(message.message))

	stand_down = self.stand_down
	if stand_down is None:
		self.abort()
		return CLEARING
	seconds = spread_out(stand_down)

	def restart(self, value, args):
		role_name = args.role_name
		if self.role_name:
			a = self.create(self.object_type, *self.args, role_name=role_name, **self.settings)
		else:
			a = self.create(self.object_type, *self.args, **self.settings)
		self.assign(a, role_name)
		self.idle_object.append(a)

	# Run a no-op with the desired timeout.
	a = self.create(Delay, seconds=seconds)
	self.on_return(a, restart, role_name=d)
	return SPOOLING

def ObjectSpool_SPOOLING_Stop(self, message):
	if self.abort():
		return CLEARING
	self.complete(Aborted())

def ObjectSpool_CLEARING_Returned(self, message):
	d = self.debrief()
	if self.working():
		return CLEARING
	self.complete(Aborted())

OBJECT_SPOOL_DISPATCH = {
	INITIAL: (
		(Start,),
		()
	),
	SPOOLING: (
		(JoinSpool, LeaveSpool, Unknown, Returned, Stop),
		()
	),
	CLEARING: (
		(Returned,),
		()
	),
}

bind(ObjectSpool, OBJECT_SPOOL_DISPATCH, thread='object-spool')



# test_api.py

# A basic technique for managing ports used during
# unit testing. All ports start at the number below.
# Each test module that needs to use network ports
# is assigned a 25-port block, i.e. the 4th module
# has this line at the top;
# TEST_PORT = TEST_PORT_START + 75
# and then each unit test uses the notation;
# listen(self, HostPort('127...', TEST_PORT + 1))
# where the additional index refers to a slot
# within the block of 25.

__all__ = [
	'TEST_PORT_START',
]

# Somewhere in the dynamic, private, ephemeral space.
TEST_PORT_START = 55195



# test_function_1.py
import random

random.seed()

def texture(x: int=8, y: int=8) -> list[list[float]]:
	table = []
	for r in range(y):
		row = [None] * x
		table.append(row)
		for c in range(x):
			row[c] = random.random()

	return table



# Author: Scott Woods <kipjak.baker@gmail.com>
# MIT License
#
# Copyright (c) 2025 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Making an exclusive claim for a persisted name.

"""
__docformat__ = 'restructuredtext'

from .virtual_runtime import *
from .virtual_point import *
from .point_runtime import *
from .virtual_memory import *
from .message_memory import *
from .routine_point import *
from .file_object import *
from .bind_type import *

__all__ = [
	'lock_file',
	'unlock_file',
	'LockUp',
	'LockedOut',
	'head_lock',
]

#
#
class LockedOut(object):
	"""Other process already working in same space."""
	def __init__(self, path: str=None, pid: int=None, parent_pid: int=None):
		self.path = path
		self.pid = pid
		self.parent_pid = parent_pid

	def __str__(self):
		if self.path is None or self.pid is None or self.parent_pid is None:
			return 'locked out and missing details'
		s = f'locked out of "{self.path}" by <{self.pid}>({self.parent_pid})'
		return s

bind_message(LockedOut, copy_before_sending=False)

# Posix based file locking (Linux, Ubuntu, MacOS, etc.)
# Only allows locking on writable files, might cause
# strange results for reading.

try:
	import fcntl
	import os

	def lock_file(f):
		# Will throw OSError if file is not writable
		# or someone else has already locked it.
		fcntl.lockf(f.fileno(), fcntl.LOCK_EX | fcntl.LOCK_NB)

	def unlock_file(f):
		fcntl.lockf(f.fileno(), fcntl.LOCK_UN)

except ModuleNotFoundError:
	import msvcrt
	import os

	def file_size(f):
		return os.path.getsize(os.path.realpath(f.name))

	def lock_file(f):
		msvcrt.locking(f.fileno(), msvcrt.LK_RLCK, file_size(f))

	def unlock_file(f):
		msvcrt.locking(f.fileno(), msvcrt.LK_UNLCK, file_size(f))


# Class for ensuring that all file operations are atomic, treat
# initialization like a standard call to 'open' that happens to be atomic.
# This file opener *must* be used in a "with" block.


class LockUp:
	def __init__(self, path, *args, **kwargs):
		self.file = open(path, *args, **kwargs)
		lock_file(self.file)
		f = File(path, UserDefined(LockedOut))
		lo = LockedOut()
		lo.path = path
		lo.pid = os.getpid()
		lo.parent_pid = os.getppid()
		f.store(lo)

	def __enter__(self, *args, **kwargs):
		return self.file

	def __exit__(self, exc_type=None, exc_value=None, traceback=None):
		# Flush to make sure all buffered contents are written to file.
		self.file.flush()
		os.fsync(self.file.fileno())

		# Release the lock on the file.
		unlock_file(self.file)
		self.file.close()

		# Handle exceptions that may have come up during execution, by
		# default any exceptions are raised to the user.
		if (exc_type is not None):
			return False
		return True

#
#
def head_lock(self, path, name, ):
	lock = os.path.join(path, name)
	try:
		with LockUp(lock, "w", ) as k:
			self.send(Ready(), self.parent_address)
			self.select(Stop)
		return None
	except OSError:
		# Address a not needed in completion code.
		f = File(lock, UserDefined(LockedOut))
		lo = f.recover()
		return lo

bind_routine(head_lock)



# Author: Scott Woods <kipjak.baker@gmail.com>
# MIT License
#
# Copyright (c) 2025 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Primitives and data structures supporting a thread-safe, object space.

"""
__docformat__ = 'restructuredtext'

import threading

from .virtual_memory import *
from .message_pump import *

__all__ = [
	'Completion',
	'create_an_object',
	'destroy_an_object',
	'find_object',
	'OpenAddress',
	'send_a_message',
	'set_queue',
	'get_queue',
	'get_queue_address',
]

# Mechanism for jumping out of non-thread objects, e.g. machines.
class Completion(Exception):
	def __init__(self, message=None):
		self.message = message

# THE BIG BANG STARTS HERE.
# Core async object management.
# Safe serial ids, type library, active objects and messaging.
serial_number = 1
serial_lock = threading.RLock()

def get_next_id():
	"""Thread-safe increment of serial id, return the original value.

	Used during object creation for assignment of unique ids.

	Returns:

	a previously unused, unique kipjak address, suitable for use as a
	send destination.
	"""
	global serial_number, serial_lock
	try:
		serial_lock.acquire()
		next = serial_number
		serial_number += 1
	finally:
		serial_lock.release()
	return (next,)

#
#
object_map  = {}
object_lock = threading.RLock()

def create_an_object(object_type, object_ending, parent_address, args, kw_args):
	"""
	Thread-safe construction of an async object, e.g. queue, thread or
	point. Returns address, object tuple.

	The object is instantiated using the args+kw_args supplied, if any. Then
	the object is pre-loaded with kipjak internal information, before
	returning to caller.

	Parameters:

	- `object_type`: a class, the application class that derives from a
	combination of kipjak base classes.
	- `parent_address`: an kipjak address, the address of the object that
	is creating the new object.
	- `args`: tuple of values, the args forwarded on to the object
	__init__ call using `*args`.
	- `kw_args`: dict of named values, the named args forwarded on to the
	object __init__ call using `**kw_args`.

	Returns:

	a 2-tuple of an kipjak address and an async object instance. The
	former being the unique id for the latter.
	"""
	global object_map, object_lock
	address = get_next_id()
	try:
		created = object_type(*args, **kw_args)
		created.object_ending = object_ending
	except TypeError as e:
		raise RuntimeError(str(e))

	# Some jiggery to resolve the queue for this object. Need the object
	# handle for the buffering and the address for processing of messages.
	# Cant deliver to a queue that may have been destroyed - must always
	# perform lookup during send.
	if issubclass(object_type, Pump):
		assigned_queue = created
	else:
		assigned_queue = get_queue(object_type)

	created.object_address = address
	created.assigned_queue = assigned_queue
	if assigned_queue is None:
		created.queue_address = NO_SUCH_ADDRESS
	else:
		created.queue_address = assigned_queue.object_address
	created.parent_address = parent_address

	try:
		object_lock.acquire()
		object_map[address[-1]] = created
	finally:
		object_lock.release()
	return address, created

def destroy_an_object(address):
	"""
	Thread-safe destruction of an async object. No return.

	Uses the specified address to remove all trace of the object
	from kipjak.

	Returns:

	Nothing.
	"""
	global object_map, object_lock
	try:
		object_lock.acquire()
		try:
			del object_map[address[-1]]
		except KeyError:
			pass
	finally:
		object_lock.release()

def find_object(address):
	"""
	Thread-safe object lookup by address.

	Uses the specified address to retrieve the associated, actual
	object.  This call is dangerous to anyone other than the thread
	object assigned to handle message processing. If the wrong thread
	calls this function the returned object could be deleted from under
	it.

	Parameters:

	- `address`: a kipjak address, the address of the object to be
	retrieved.

	Returns:

	an async object instance or None, if the underlying object no longer
	exists.
	"""
	global object_map, object_lock
	try:
		object_lock.acquire()
		try:
			matched = object_map[address[-1]]
		except KeyError:
			return None
	finally:
		object_lock.release()
	return matched

# Class for safe access to the object underlying an address.
# Used with real caution.
class OpenAddress:
	def __init__(self, address, *args, **kwargs):
		self.object_address = address

	def __enter__(self, *args, **kwargs):
		global object_map, object_lock
		object_lock.acquire()
		try:
			return object_map[self.object_address[-1]]
		except KeyError:
			return None

	def __exit__(self, exc_type=None, exc_value=None, traceback=None):
		global object_map, object_lock
		object_lock.release()
		if (exc_type is not None):
			return False
		return True

#
#
def send_a_message(message, to_address, return_address):
	"""
	Thread-safe transfer of a message between the given addresses. No
	return value and no failures.

	Abstractly messages are sent to a receiving async object.
	Technically that involves up to two lookups 1) find the target
	object and 2) find the queue assigned to that target object. In the
	special case where the target object is its own assigned queue then
	the latter is not necessary.

	The message is appended to the retrieved queue. There is the
	expectation that the assigned thread will take it from the queue and
	either consume it itself (i.e. its a custom object or sync object)
	or present it to one of the auto-dispatching machines (i.e.
	Stateless or StateMachine).

	If either of the two lookups fails - the destination object or the
	assigned queue object do not exist - the send fails silently. This
	is one of the code sites that implements the fall-on-the-floor
	handling of any message that cant be sent. See also the dispatch
	machines where there is no appropriate handler for a certain message
	type.

	Parameters:

	- `message`: a plain, old data object, the Message-based object to
	be transferred.
	- `to_address`: a kipjak address, the async object that will receive
	the message.
	- `return_address`: a kipjak address, the async object that called
	`send`.

	Returns:

	Nothing.
	"""
	global object_map, object_lock
	try:
		object_lock.acquire()
		try:
			to_object = object_map[to_address[-1]]
			if to_object.queue_address[-1] == to_address[-1]:	  # Dont need to lookup self.
				queue_object = to_object
			else:
				queue_object = object_map[to_object.queue_address[-1]]
			queue_object.put([message, to_address, return_address])
		except KeyError:			# No-such-address is not an error.
			pass
	finally:
		object_lock.release()

#
#
type_map = {}
type_lock = threading.RLock()

def set_queue(object_type, queue_address):
	"""
	"""
	global type_map, type_lock
	try:
		type_lock.acquire()
		try:
			queue = object_map[queue_address[-1]]
		except KeyError:
			return
		type_map[object_type] = queue
	finally:
		type_lock.release()

def get_queue(object_type):
	"""
	"""
	global type_map, type_lock
	assigned_queue = None

	try:
		type_lock.acquire()
		try:
			assigned_queue = type_map[object_type]  # Registered queue.
		except KeyError:
			assigned_queue = type_map[None]
	finally:
		type_lock.release()
	return assigned_queue

def get_queue_address(object_type):
	"""
	"""
	global type_map, type_lock
	assigned_queue = None

	try:
		type_lock.acquire()
		try:
			assigned_queue = type_map[object_type]  # Registered queue.
		except KeyError:
			assigned_queue = type_map[None]
		queue_address = assigned_queue.object_address
	finally:
		type_lock.release()
	return queue_address



# test_echo.py
import kipjak as kj

def main(self, requested_ipp: kj.HostPort=None):
	requested_ipp = requested_ipp or kj.HostPort('127.0.0.1', 5010)

	self.console(requested_ipp=requested_ipp)

	# Establish the network listen.
	kj.listen(self, requested_ipp=requested_ipp)
	m, i = self.select(kj.Listening, kj.NotListening, kj.Stop)
	if i == 1:
		return m
	if i == 2:
		return kj.Aborted()

	# Errors, sessions and inbound client messages.
	while True:
		m, i = self.select(kj.NotListening,		# Listen failed.
			kj.Accepted, kj.Closed,				# Session notifications.
			kj.Stop,							# Intervention.
			kj.Unknown)							# An inbound message.

		if i == 0:				# Terminate with the fault.
			break
		elif i in (1, 2):		# Ignore.
			continue
		elif i == 3:			# Terminate as requested.
			m = kj.Aborted()
			break

		rt = type(self.received_type)
		self.console(typ=rt)
		if rt == kj.UserDefined:
			self.console(element=self.received_type.element)
		c = kj.cast_to(m, self.received_type)	# Send the message back over the connection.
		self.reply(c)

	return m

kj.bind(main, entry_point=(kj.Ack, kj.Enquiry))

if __name__ == '__main__':
	kj.create(main)



# test_server.py
import kipjak as kj

# A bare-bones implementation of a traditional network server that
# demonstrates the different function-calling options.

def thing(self, publish_as, subscribe_to, seconds=None):
	kj.publish(self, publish_as, scope=kj.ScopeOfDirectory.PROCESS)
	kj.subscribe(self, subscribe_to, scope=kj.ScopeOfDirectory.PROCESS)

	if seconds:
		self.start(kj.T1, seconds)

	published, i = self.select(kj.Published, saving=kj.Subscribed)	# Published...
	subscribed, i = self.select(kj.Subscribed, saving=kj.Published)

	while True:
		m = self.input()
		if not isinstance(m, (kj.Available, kj.Delivered, kj.Dropped)):
			break

	kj.clear_published(self, published)
	self.input()							# PublishedCleared

	kj.clear_subscribed(self, subscribed)
	self.input()							# SubscribedCleared

	return None

kj.bind(thing)

def abc(self, server_address: kj.HostPort=None, flooding: int=64, soaking: int=100):
	a = self.create(thing, "a", "c")
	b = self.create(thing, "b", "a", seconds=5.0)
	c = self.create(thing, "c", "b")

	self.assign(a, 0)
	self.assign(b, 1)
	self.assign(c, 2)

	m = self.input()
	while True:
		if isinstance(m, kj.Stop):
			self.abort()

		elif isinstance(m, kj.Returned):
			self.debrief()
			m = self.input()
			continue

		else:
			continue

		while self.working():
			self.input()
			self.debrief()

		return None

	return None

kj.bind(abc)	# Register with the framework.

if __name__ == '__main__':	# Process entry-point.
	kj.create(abc)



# Author: Scott Woods <kipjak.baker@gmail.com>
# MIT License
#
# Copyright (c) 2025 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Manage folders of files containing encodings of application data.

The :class:`~.Folder` class provides for one-time description of folder location and
contents. The description is inherited by any child objects (i.e. ``Files``
and sub-``Folders``) created, and by operations within the folder.

The class also provides for persistence of maps. Rather than using a file
to store an entire map, the :class:`~.Folder` class can be used to store the
map on a one-file-per-entry basis.
"""

__docformat__ = 'restructuredtext'

# .. autoclass:: Folder
#   :members:
#   :no-undoc-members:

__all__ = [
	'Folder',
	'remove_folder',
	'remove_contents',
	'shape_of_folder',
]

import os
import errno
import re as regex

from .virtual_memory import *
from .message_memory import *
from .make_message import *
from .convert_type import *
from .virtual_codec import *
from .json_codec import *
from .file_object import *

#
#
def remove_contents(path):
	"""Delete everything under the given folder, down."""

	for f in os.listdir(path):
		p = os.path.join(path, f)
		if os.path.isdir(p):
			remove_folder(p)
		elif os.path.isfile(p):
			os.remove(p)

def remove_folder(path):
	"""Delete everything under the given folder, and then the folder."""

	remove_contents(path)

	try:
		os.rmdir(path)
	except OSError as e:
		if e.errno == errno.ENOENT:
			return

def shape_of_folder(path):
	"""Walk the given folder, acumulating folders, files and bytes (3-tuple)."""
	folders, files, bytes = 0, 0, 0

	p = path
	for f in os.listdir(path):
		p = os.path.join(path, f)
		if os.path.isdir(p):
			folders += 1
			fo, fi, by = shape_of_folder(p)
			folders += fo
			files += fi
			bytes += by
		elif os.path.isfile(p):
			s = os.stat(p)
			files += 1
			bytes += s.st_size

	return folders, files, bytes

#
#
class Folder(object):
	"""Create and manage a folder of application encodings.

	:param path: the location in the filesystem
	:param tip: type expression for the content
	:type tip: :ref:`tip<type-reference>`
	:param re: formal, regular expression description of expected file names
	:param encoding: selection of representation, defaults to ``CodecJson``
	:type encoding: class
	:param pretty_format: generate human-readable file contents, defaults to ``True``
	:param decorate_names: auto-append an encoding-dependent extension to the file name, defaults to ``True``
	:param keys_names: a key-composer function and a name-composer function
	:type keys_names: 2-tuple of functions
	:param make_absolute: expand a relative path to be an absolute location, defaults to ``True``
	:param auto_create: create folders as necessary, defaults to ``True``
	"""

	def __init__(self, path: str=None,
			tip=None, re: str=None, encoding=None,
			pretty_format: bool=True, decorate_names: bool=True,
			create_default: bool=False, keys_names=None,
			make_absolute: bool=False, auto_create: bool=True):
		"""Construct a Folder instance."""
		path = path or '.'
		if make_absolute:
			path = os.path.abspath(path)
		self.path = path

		if re is None:
			self.re = None
		else:
			self.re = regex.compile(re)

		if tip is None:
			self.file_type = None
		elif isinstance(tip, Portable):
			self.file_type = tip
		elif hasattr(tip, '__art__'):
			self.file_type = UserDefined(tip)
		else:
			self.file_type = lookup_type(tip)

		self.encoding = encoding or CodecJson
		self.pretty_format = pretty_format
		self.decorate_names = decorate_names
		self.create_default = create_default
		self.keys_names = keys_names
		self.auto_create = auto_create

		if not auto_create:
			return
		try:
			os.makedirs(self.path)
		except OSError as e:
			if e.errno == errno.EEXIST:
				return

	def folder(self, name: str, tip=None, re: str=None, encoding=None,
			pretty_format: bool=None, decorate_names: bool=None, create_default: bool=None,
			auto_create: bool=None, keys_names=None):
		"""Create a new :class:`~.Folder` object representing a sub-folder at the current location.

		:param path: the name to be added to the saved ``path``
		:param tip: type expression for the content
		:type tip: :ref:`tip<type-reference>`
		:param re: formal, regular expression description of expected file names
		:param encoding: selection of representation, defaults to ``CodecJson``
		:type encoding: class
		:param pretty_format: generate human-readable file contents, defaults to ``True``
		:param decorate_names: auto-append an encoding-dependent extension to the file name, defaults to ``True``
		:param keys_names: a key-composer function and a name-composer function
		:type keys_names: 2-tuple of functions
		:param make_absolute: expand a relative path to be an absolute location, defaults to ``True``
		:param auto_create: create folders as necessary, defaults to ``None``
		:return: a new location in the filesystem
		:rtype: Folder
		"""
		tip = tip or self.file_type
		if re is None:
			self.re = None
		else:
			self.re = regex.compile(re)
		encoding = encoding or self.encoding
		if pretty_format is None: pretty_format = self.pretty_format
		if decorate_names is None: decorate_names = self.decorate_names
		if create_default is None: create_default = self.create_default
		if auto_create is None: auto_create = self.auto_create
		keys_names = keys_names or self.keys_names

		path = os.path.join(self.path, name)
		return Folder(path, re=re, tip=tip, encoding=encoding,
			pretty_format=pretty_format, decorate_names=decorate_names, create_default=create_default,
			keys_names=keys_names, make_absolute=False, auto_create=auto_create)

	def file(self, name: str, tip=None, encoding=None,
			pretty_format: bool=None, decorate_names: bool=None, create_default: bool=None):
		"""Create a new :class:`~.File` object representing a file at the current location.

		:param name: the name to be added to the saved ``path``
		:param tip: type expression for the content
		:type tip: :ref:`tip<type-reference>`
		:param encoding: selection of representation, defaults to ``CodecJson``
		:type encoding: class
		:param pretty_format: generate human-readable file contents, defaults to ``True``
		:param decorate_names: auto-append an encoding-dependent extension to the file name, defaults to ``True``
		:param create_default: return default instance on file not found, defaults to ``False``
		:return: a new file in the filesystem
		:rtype: File
		"""
		tip = tip or self.file_type
		encoding = encoding or self.encoding
		if pretty_format is None: pretty_format = self.pretty_format
		if decorate_names is None: decorate_names = self.decorate_names
		if create_default is None: create_default = self.create_default

		path = os.path.join(self.path, name)	# Let the I/O operation decorate.
		return File(path, tip, encoding=encoding,
			pretty_format=pretty_format, decorate_names=decorate_names, create_default=create_default)

	def matching(self):
		"""Scan for files in the folder.

		:return: a sequence of filenames matching the :class:`~.Folder` criteria.
		:rtype: str
		"""
		re = self.re
		decorate_names = self.decorate_names
		extension = '.%s' % (self.encoding.EXTENSION,)
		for f in os.listdir(self.path):
			m = None
			p = os.path.join(self.path, f)
			if not os.path.isfile(p):
				continue
			if decorate_names:
				b, e = os.path.splitext(f)
				if e != extension:
					continue
				f = b
			if re:
				m = re.fullmatch(f)
				if not m:
					continue
			yield f

	def each(self):
		"""Process the files in the folder.

		:return: a sequence of :class:`~.File` objects matching the :class:`~.Folder` criteria.
		:rtype: File
		"""
		# Get a fresh image of folder/slice.
		# Use a snapshot for iteration to avoid
		# complications arising from changes to the folder.
		matched = [f for f in self.matching()]
		# Visit each named file.
		# Yield a file object, ready for I/O.
		for f in matched:
			yield self.file(f, tip=self.file_type)

	def store(self, values: dict):
		"""Store a ``dict`` of values as files in the folder.

		:param values: a collection of application values
		"""
		# Get a fresh image of folder/slice.
		matched = set(self.matching())
		stored = set()
		for k, v in values.items():
			name = self.name(v)
			io = self.file(name, tip=self.file_type)
			io.store(v)
			stored.add(name)
		# Clean out files that look like they
		# have been previously written but are
		# no longer in the map.
		matched -= stored
		for m in matched:
			self.erase(m)

	def recover(self):
		"""Recover application values from the files in the folder.

		A generator function that yields a sequence of tuples that
		allow the caller to process an entire folder with a clean loop.

		All arguments are forwarded to :meth:`~.file_object.recover`.

		The return value includes the version of the main decoded object, or None
		if the encoding and decoding applications are at the same version. This value is
		the mechanism by which applications can select different code-paths in support of
		older versions of encoded materials.

		:return: a sequence of 2-tuples, 0) key and 1) the value
		:rtype: tuple
		"""
		# Get a fresh image of folder/slice.
		matched = [f for f in self.matching()]
		# Visit each named file.
		# Yield the key, message tuple.
		for f in matched:
			io = self.file(f, tip=self.file_type)
			r = io.recover()
			if self.keys_names is None:
				k = None
			else:
				k = self.key(r)
			yield k, r

	def add(self, values: dict, item):
		"""Add a value, both to a ``dict`` of values and as a file in the folder.

		:param values: a collection of application values
		:param item: the value to be added
		:type item: :ref:`tip<type-reference>`
		"""
		keys_names = self.keys_names
		if keys_names is None:
			raise ValueError(f'key/name functions not set for "{self.path}" (add)')

		key = keys_names[0](item)
		name = keys_names[1](item)

		io = self.file(name, tip=self.file_type)
		if key in values:
			raise ValueError(f'name "{io.name}" already present (add)')
		io.store(item)
		values[key] = item

	def update(self, values: dict, item):
		"""Update a value, both in a ``dict`` of values and as a file in the folder.

		:param values: a collection of application values
		:param item: the value to be updated
		:type item: :ref:`tip<type-reference>`
		"""
		keys_names = self.keys_names
		if keys_names is None:
			raise ValueError(f'key/name functions not set for "{self.path}" (update)')

		key = keys_names[0](item)
		name = keys_names[1](item)

		io = self.file(name, tip=self.file_type)
		if key not in values:
			raise ValueError(f'name "{io.name}" not an existing entry (update)')

		io.store(item)
		values[key] = item

	def remove(self, values: dict, item):
		"""Remove a value, both from a ``dict`` of values and as a file in the folder.

		:param values: a collection of application values
		:param item: the value to be removed
		:type item: :ref:`tip<type-reference>`
		"""
		keys_names = self.keys_names
		if keys_names is None:
			raise ValueError(f'key/name functions not set for "{self.path}" (remove)')
		key = keys_names[0](item)
		name = keys_names[1](item)

		self.erase(name)
		del values[key]

	def clear(self, values: dict):
		"""Remove all values, both from a ``dict`` of values and as files in the folder.

		:param values: a collection of application values
		"""
		# Brute force. Delete any candidates from
		# the folder and dump everything from the dict.
		matched = [f for f in self.matching()]
		for removing in matched:
			self.erase(removing)
		values.clear()

	def erase(self, name: str):
		"""Delete the named file from the folder.

		:param name: a name of a file
		"""
		path = os.path.join(self.path, name)
		name = path
		if self.decorate_names:
			name = '%s.%s' % (path, self.encoding.EXTENSION)
		if os.path.isfile(name):
			os.remove(name)
			return True
		elif os.path.isdir(path):
			remove_folder(path)
			return True
		return False

	def exists(self, name: str=None):
		"""Detect the named file, within the folder.

		:param name: a name of a file
		:return: does the file exist
		:rtype: bool
		"""
		if name is None:
			return os.path.isdir(self.path)

		path = os.path.join(self.path, name)
		name = path
		if self.decorate_names:
			name = '%s.%s' % (path, self.encoding.EXTENSION)
		if os.path.isfile(name):
			return True
		elif os.path.isdir(path):
			return True
		return False

	def key(self, item):
		"""Generate the stable key for a given application value.

		:param item: an application value
		:type item: :ref:`tip<type-reference>`
		:return: the key
		:rtype: folder dependent
		"""
		keys_names = self.keys_names
		if keys_names is None:
			raise ValueError(f'key/name functions not set for "{self.path}" (key)')
		return keys_names[0](item)

	def name(self, item):
		"""Generate the stable filename for a given application value.

		:param item: an application value
		:type item: :ref:`tip<type-reference>`
		:return: the filename
		:rtype: str
		"""
		keys_names = self.keys_names
		if keys_names is None:
			raise ValueError(f'key/name functions not set for "{self.path}" (name)')
		return keys_names[1](item)



# Author: Scott Woods <kipjak.baker@gmail.com>
# MIT License
#
# Copyright (c) 2025 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Self-maintaining storage as a FIFO of files.

Creates and maintains a series of files where each
file contains a time-slice of logs. New lines are added
to the end of the latest file until that reaches a
maximum size. At which point a new file is "tacked on" at
the end. Oldest files are deleted as necessary.
"""
__docformat__ = 'restructuredtext'

import os
import time
import re

from .folder_object import *
from .convert_memory import *
from collections import deque

__all__ = [
	'LINES_IN_FILE',
	'FILES_IN_FOLDER',
	'RollingLog',
	'read_log',
	'rewind_log',
]

#
#
YMDTHMSF = '([0-9]{4})h([0-9]{2})h([0-9]{2})T([0-9]{2})c([0-9]{2})c([0-9]{2})p([0-9]{3})'
naming_convention = re.compile(YMDTHMSF)

LINES_IN_FILE = 16384
FILES_IN_FOLDER = 512

class RollingLog(object):
	"""
	"""
	def __init__(self, path, lines_in_file=None, files_in_folder=None):
		self.path = path
		self.folder = Folder(path, re=YMDTHMSF, decorate_names=False)
		self.lines_in_file = lines_in_file or LINES_IN_FILE
		self.files_in_folder = files_in_folder or FILES_IN_FOLDER
		self.lines = 0

		flat = []
		for f in self.folder.matching():
			iso = f.replace('h', '-')
			iso = iso.replace('c', ':')
			iso = iso.replace('p', '.')
			a = [
				f,
				text_to_world(iso),
			]
			flat.append(a)

		flat.sort(key=lambda m: m[1])
		self.manifest = deque(flat)
		self.opened, _ = self.open_file(time.time())

	def log_time(self, t):
		second = time.strftime('%Y-%m-%dT%H:%M:%S', time.gmtime(t))
		fraction = '%.3f' % (t,)
		fraction = fraction[-3:]
		lt = '%s.%s' % (second, fraction)
		return lt

	def open_file(self, t):
		lt = self.log_time(t)
		hcp = lt.replace('-', 'h')
		hcp = hcp.replace(':', 'c')
		hcp = hcp.replace('.', 'p')
		path = os.path.join(self.folder.path, hcp)
		f = open(path, 'w')
		self.manifest.append([path, t])
		while len(self.manifest) > self.files_in_folder:
			a = self.manifest.popleft()
			os.remove(a[0])
		self.lines = 0
		return f, lt

	def close_file(self, opened):
		opened.close()

	def __call__(self, log):
		"""
		"""
		if self.lines < self.lines_in_file:
			lt = self.log_time(log.stamp)
		else:
			self.close_file(self.opened)
			self.opened, lt = self.open_file(log.stamp)

		name = log.name.split('.')[-1]
		state = log.state
		if state is None:
			line = '%s %s <%08x>%s - %s\n' % (lt, log.tag.value, log.address[-1], name, log.text)
		else:
			line = '%s %s <%08x>%s[%s] - %s\n' % (lt, log.tag.value, log.address[-1], name, state, log.text)
		self.opened.write(line)
		self.opened.flush()

		self.lines += 1
		return line
	#def __close__(self):
	#	"Proper termination of file-based logging."
	#	pass

def read_log(logs, begin, end, count):
	'''Coroutine that accepts a log folder, range and yields lines.'''

	# Get the collection of files and their
	# timestamps.
	folder = Folder(logs.path, re=YMDTHMSF, decorate_names=False)
	rolling = []
	for f in folder.matching():
		iso = f.replace('h', '-')
		iso = iso.replace('c', ':')
		iso = iso.replace('p', '.')
		d = text_to_world(iso)
		a = [
			os.path.join(folder.path, f),
			d,
		]
		rolling.append(a)

	if len(rolling) < 1:	   # Early exit if nothing there.
		return
	rolling.sort(key=lambda m: m[1])		# Put them in order.

	if end is not None and end < rolling[0][1]:	 # Timeframe before all records.
		return

	# Internal coroutines to handle file and line
	# scanning.
	def get_file():
		# Slide up to the file with the "nearest"
		# timestamp, i.e. the one before the file
		# that is after.
		n = len(rolling)
		if begin < rolling[0][1]:
			i = 0
		elif begin >= rolling[-1][1]:
			yield rolling[-1][0]
			return
		else:
			for i in range(n - 1):
				if rolling[i + 1][1] > begin:
					break

		# Yield the sequence of files starting
		# at i.
		for j in range(i, len(rolling)):
			yield rolling[j][0]

	def get_line(r):
		# Open and close the presented file. Yield
		# each line from the file along with the
		# timestamp in usable form.
		with open(r, 'r') as f:
			for line in f:
				# convert stamp 2023-01-03T19:23:51
				i = line.index(' ')
				t = line[:i]
				d = text_to_world(t)
				yield d, line

	# 3 variants on the querying of a log,
	# 1) from <begin> to <end>,
	# 2) from <begin> for <count> and
	# 3) from <begin> to end-of-log.
	if end is not None:
		for r in get_file():
			for d, l in get_line(r):
				if d < begin:
					continue
				if d < end:
					yield d, l
				else:
					return
	elif count is not None:
		for r in get_file():
			for d, l in get_line(r):
				if d < begin:
					continue
				if count == 0:
					return
				count -= 1
				yield d, l
	else:
		for r in get_file():
			for d, l in get_line(r):
				if d < begin:
					continue
				yield d, l

def rewind_log(logs, tail, end, count):
	'''Coroutine that accepts a log folder, range and yields lines.'''

	# Get the collection of files and their
	# timestamps.
	folder = Folder(logs.path, re=YMDTHMSF, decorate_names=False)
	rolling = []
	for f in folder.matching():
		iso = f.replace('h', '-')
		iso = iso.replace('c', ':')
		iso = iso.replace('p', '.')
		d = text_to_world(iso)
		a = [
			os.path.join(folder.path, f),
			d,
		]
		rolling.append(a)

	if len(rolling) < 1:	   # Early exit if nothing there.
		return
	rolling.sort(key=lambda m: m[1], reverse=True)

	if end is not None and end < rolling[0][1]:	 # Timeframe before all records.
		return

	def get_file():
		for r in rolling:
			yield r[0]

	def get_line(r):
		with open(r, 'r') as f:
			for line in f:
				# convert stamp 2023-01-03T19:23:51
				i = line.index(' ')
				t = line[:i]
				d = text_to_world(t)
				yield d, line

	def get_ending(n):
		ending = []
		for r in get_file():
			q = deque()
			for dl in get_line(r):
				q.append(dl)
				if len(q) > n:
					q.popleft()
			ending.append(q)
			n -= len(q)
			if n == 0:
				return ending
		return ending

	ending = get_ending(tail)
	for r in reversed(ending):
		if end is not None:
			for d, l in r:
				if d < end:
					yield d, l
				else:
					return
		elif count is not None:
			for d, l in r:
				if count == 0:
					return
				count -= 1
				yield d, l
		else:
			for d, l in r:
				yield d, l
